package com.framsticks.gui.components;

import com.framsticks.params.Flags;
import com.framsticks.params.types.DecimalParam;
import com.framsticks.params.types.FloatParam;
import com.framsticks.params.types.ValueParam;
import com.framsticks.util.Layout;
import com.framsticks.util.Numbers;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.Hashtable;


@SuppressWarnings("serial")
public class SliderControl extends ValueControl {

	private JSlider slider;

	private JTextField sliderValue;

	/**
	 * Division factor used to implement float value slider.
	 */
	private final int div = 10;

	/**
	 * Old value of slider used when user press ESCAPE key.
	 */
	private int oldValue;

	public SliderControl(ValueParam valueParam) {
		super(valueParam);
		slider = new JSlider();
        slider.setEnabled(!isReadOnly());
		slider.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent changeEvent) {
				notifyOfChange();
			}
		});
		if (param instanceof DecimalParam) {
			int min = param.getMin(Integer.class);
			int max = param.getMax(Integer.class);
			slider.setMinimum(min);
			slider.setMaximum(max);
			if (param.getDef(Integer.class) != null) {
				slider.setValue(param.getDef(Integer.class));
			} else {
				slider.setValue(min);
			}
			slider.setMajorTickSpacing((int) ((max - min) / 5));
			slider.setMinorTickSpacing((int) ((max - min) / 10));
			slider.setPaintLabels(true);
			slider.setPaintTicks(true);
		}
		if (param instanceof FloatParam) {
			double min = param.getMin(Double.class) * div;
			slider.setMinimum((int) min);
			double max = param.getMax(Double.class) * div;
			slider.setMaximum((int) max);

			Hashtable<Integer, java.awt.Component> labels = new Hashtable<Integer, java.awt.Component>();
			for (int i = 0; i < 6; i++) {
				double val = 0;
				if (min < 0) {
					val = (((Math.abs(max) - min) / 5) * i) + min;
				} else {
					val = (((Math.abs(max) - Math.abs(min)) / 5) * i) + Math.abs(min);
				}
				String label = String.format("%1$.1f", (val / 10)).replace(",", ".");
				labels.put(new Integer((int) val), new JLabel(label, JLabel.CENTER));
			}
			slider.setLabelTable(labels);
			slider.setPaintLabels(true);
			slider.setMajorTickSpacing((int) ((max - Math.abs(min)) / 5));
			slider.setMinorTickSpacing((int) ((max - Math.abs(min)) / 10));
			slider.setPaintTicks(true);
			if (param.getDef(Double.class) != null) {
				double defaultValue = param.getDef(Double.class) * div;
				slider.setValue((int) defaultValue);
			}
		}
		//TODO: actually set the value
		/*
		if (property.getValue() != null) {
			setValueImpl(property.getValue());
		}
		*/



		createPanel();
	}

	/**
	 * Creates panel with description in label, slider value and slider next to description.
	 */
	private void createPanel() {
		this.setLayout(new BoxLayout(this, BoxLayout.LINE_AXIS));
		this.setAlignmentX(Box.CENTER_ALIGNMENT);
		this.setAlignmentY(Box.CENTER_ALIGNMENT);

		JPanel sliderPanel = new JPanel();
		sliderPanel.setLayout(new BoxLayout(sliderPanel, BoxLayout.LINE_AXIS));

		createSliderValue();

		JPanel sVPanel = new JPanel();
		sVPanel.setLayout(new BoxLayout(sVPanel, BoxLayout.LINE_AXIS));
		sVPanel.add(sliderValue);
		Layout.copyComponentDimensions(sVPanel, sliderValue);

		JPanel sPanel = new JPanel();
		sPanel.setLayout(new BoxLayout(sPanel, BoxLayout.LINE_AXIS));

		sliderPanel.setLayout(new BorderLayout());
		sliderPanel.add(slider, BorderLayout.CENTER);
		sliderPanel.setMaximumSize(new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE));
		sliderPanel.setMinimumSize(new Dimension(0, 60));

		sPanel.add(sVPanel);
		sPanel.add(sliderPanel);

		this.add(sPanel);
	}

	/**
	 * Creates slider value text field
	 */
	private void createSliderValue() {
		if (param instanceof DecimalParam) {
			sliderValue = new JTextField(Integer.toString(slider.getValue()));
			sliderValue.setText(Integer.toString(slider.getValue()));
		} else {
			sliderValue = new JTextField(Double.toString(slider.getValue() / div));
			sliderValue.setText(Double.toString(slider.getValue() / div));
		}
		Layout.fixComponentSize(sliderValue, new Dimension(60, Control.LINE_HEIGHT));
		sliderValue.setHorizontalAlignment(JSlider.CENTER);

		if (param.hasFlag(Flags.READONLY)) {
			sliderValue.setEnabled(false);
		}

		if (param.getDef(Object.class) != null) {
			sliderValue.setText(param.getDef(Object.class).toString());
		}

		slider.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				if (param instanceof DecimalParam) {
					sliderValue.setText(Integer.toString(slider.getValue()));
				}
				if (param instanceof FloatParam) {
					float val = (float) slider.getValue() / (float) div;
					sliderValue.setText(Double.toString(val));
				}
			}
		});

		sliderValue.addFocusListener(new FocusListener() {

			public void focusGained(FocusEvent e) {
				oldValue = slider.getValue();
			}

			public void focusLost(FocusEvent e) {
				applyNewValue();
			}

		});

		sliderValue.addKeyListener(new KeyListener() {

			public void keyPressed(KeyEvent e) {
			}

			public void keyReleased(KeyEvent e) {
				if (e.getKeyCode() == KeyEvent.VK_ENTER) {
					applyNewValue();
				}
				if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
					if (param instanceof DecimalParam) {
						sliderValue.setText(Integer.toString(oldValue));
					}
					if (param instanceof FloatParam) {
						sliderValue.setText(Double.toString((float) oldValue / (float) div));
					}
					sliderValue.setBackground(Color.WHITE);
					sliderValue.invalidate();
				}
			}

			public void keyTyped(KeyEvent e) {
			}

		});
	}

	/**
	 * Applies new Value of slider if typed value is correct
	 */
	private void applyNewValue() {
		try {
			if (param instanceof DecimalParam) {
				int min = param.getMin(Integer.class);
				int max = param.getMax(Integer.class);

				int val = Integer.parseInt(sliderValue.getText());
				if (val < min) {
					val = min;
				}
				if (val > max) {
					val = max;
				}
				slider.setValue(val);
				sliderValue.setBackground(Color.WHITE);
			}
			if (param instanceof FloatParam) {
				double val = Double.parseDouble(sliderValue.getText());
				int min = param.getMin(Double.class).intValue();
				int max = param.getMax(Double.class).intValue();

				int intVal = (int) (val * div);
				min = min * div;
				max = max * div;
				if (intVal < min) {
					intVal = min;
				}
				if (intVal > max) {
					intVal = max;
				}
				slider.setValue(intVal);
				sliderValue.setBackground(Color.WHITE);
			}
		} catch (NumberFormatException e) {
			sliderValue.setText(getStringValue());
		}
	}



	private String getStringValue() {
		if (param instanceof DecimalParam) {
			return Integer.toString((Integer)getValue());
		}
		if (param instanceof FloatParam) {
			return Double.toString((Double)getValue());
		}
		return null;
	}

	@Override
	public void setValueImpl(Object value) {
		if (value == null) {
			return;
		}
		if (param instanceof DecimalParam) {
			slider.setValue(Numbers.cast(value, Integer.class));
		} else if (param instanceof FloatParam) {
			slider.setValue((int) ((Numbers.cast(value, Double.class)) * div));
		}
	}

	@Override
	public Object getValue() {
		if (param instanceof DecimalParam) {
			return slider.getValue();
		}
		if (param instanceof FloatParam) {
			return (double) slider.getValue() / (double) div;
		}
		return null;
	}


}
