package com.framsticks.params;

import com.framsticks.params.types.*;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Arrays;

/**
 * The class ParamBuilder helps building Param objects.
 * 
 * @author Mateusz Jarus <name.surname@gmail.com> (please replace name and
 *         surname with my personal data)
 *
 * @author Piotr Śniegowski
 */

public class ParamBuilder {
	private final static Logger LOGGER = Logger.getLogger(ParamBuilder.class.getName());

	private static final String ID_FIELD = "id";
	private static final String NAME_FIELD = "name";
	private static final String HELP_FIELD = "help";
	private static final String GROUP_FIELD = "group";
	private static final String TYPE_FIELD = "type";
	private static final String FLAGS_FIELD = "flags";

	/** The parameter getId. */
	private String id;

	/**
	 * The parameter internal getId. It's set by a user to use user's own getId in
	 * code
	 */
	private String internalId;

	/** The number of group, that parameter belongs to. */
	private Integer group = 0;

	/** The getFlags stored as a bit sum. */
	private Integer flags = 0;

	/** The parameter getName. */
	private String name;

	/** The getHelp (description) concerning parameter. */
	private String help;

	/** The getType of parameter. */
	@SuppressWarnings("rawtypes")
	private Class paramType;

	private Param param;

	/**
	 * Build Param based on provided data.
	 * 
	 * @return Param object
	 * @throws Exception
	 *             when Param getType is not defined
	 */
	public Param build()  {
        assert param != null;
		param.id = id;
		param.name = name;
		param.help = help;
		param.group = group;
		param.flags = flags;
		param.internalId = internalId;

		return param;
	}

	public ParamBuilder setId(String id) {
		this.id = id;
		return this;
	}

	public <T extends Param> ParamBuilder setType(Class<T> type) {
		this.paramType = type;
		try {
			this.param = (Param)paramType.newInstance();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		return this;
	}

	public <T extends Param> ParamBuilder setType(T type) {
		this.paramType = type.getClass();
		this.param = type;
		return this;
	}


	public ParamBuilder setInternalId(String internalId) {
		this.internalId = internalId;
		return this;
	}

	public ParamBuilder setGroup(Integer group) {
		this.group = group;
		return this;
	}

	public ParamBuilder setFlags(Integer flags) {
		this.flags = flags;
		return this;
	}

	public ParamBuilder setName(String name) {
		this.name = name;
		return this;
	}

	public ParamBuilder setType(String type) {

		LOGGER.trace("parsing type: " + type);

		String[] typeSplitted = type.split(" ");
		String first = typeSplitted[0];
		String second = typeSplitted.length > 1 ? typeSplitted[1] : null;
		String third = typeSplitted.length > 2 ? typeSplitted[2] : null;

		switch (first.charAt(0)) {
			case 'o': {
				setType(new ObjectParam(second != null ? second : first.substring(1)));
				break;
			}
			case 'p': {
				ProcedureParam procedureParam = new ProcedureParam();
				String signature = type.substring(1);
				try {
					procedureParam.parseSignature(signature);
				} catch (Exception e) {
					LOGGER.error("invalid procedure signature '" + signature + "': " + e);
				}
				setType(procedureParam);
				break;
			}
			case 'd': {

				int tildeIndex = type.indexOf("~");
				if (tildeIndex != -1) {
					setType(new EnumParam(new ArrayList<String>(Arrays.asList(type.substring(tildeIndex + 1).split("~")))));
				} else {
					if (first.length() >= 2) {
						switch (first.charAt(1)) {
							case 'b': {
								setType(BinaryParam.class);
								break;
							}
							case 'c': {
								setType(ColorParam.class);
								break;
							}
						}
					}
                    if ("0".equals(second) && "1".equals(third)) {
                        setType(BooleanParam.class);
                    }
					if (param == null) {
						setType(DecimalParam.class);
					}
				}
				if (this.param instanceof DecimalParam) {
					try {
						if (second != null) {
							this.param.min = Integer.parseInt(second);
						}

						if (third != null) {
							this.param.max = Integer.parseInt(third);
						}
					} catch (NumberFormatException e) {
						LOGGER.warn("value of min or max attribute should be an Integer");
					}
				}
				break;
			}
			case 'f': {
				setType(FloatParam.class);
				try {
					if (second != null) {
						this.param.min = Double.parseDouble(second);
					}

					if (third != null) {
						this.param.max = Double.parseDouble(third);
					}
				} catch (NumberFormatException ex) {
					LOGGER.warn("Value of min or max attribute should be Double");
				}
				break;
			}
			case 'x': {
				setType(UniversalParam.class);
				break;
			}
			case 's': {
				setType(StringParam.class);
				this.param.min = second;
				this.param.max = third;
				break;
			}
			case 'e': {
				setType(EventParam.class);
				break;
			}
			case 'l': {

				setType(third != null ? new UniqueListParam(second, third) : new ArrayListParam(second));
				break;
			}

		}
		return this;
	}

	public ParamBuilder setHelp(String help) {
		this.help = help;
		return this;
	}

	public <T> ParamBuilder setMin(T min) {
		this.param.min = min;
		return this;
	}

	public <T> ParamBuilder setMax(T max) {
		this.param.max = max;
		return this;
	}

	public <T> ParamBuilder setDef(T def) {
		this.param.def = def;
		return this;
	}

	public Param build(String line) throws Exception {
		String[] paramEntryValues = line.split(",");

		if (paramEntryValues.length == 0) {
			LOGGER.warn("field empty or wrong format (" + line
					+ ") - omitting");
			return null;
		}

		for (int i = 0; i < paramEntryValues.length; ++i) {
			paramEntryValues[i] = paramEntryValues[i].trim();
		}

		try {
			setId(paramEntryValues[0]);
			setGroup(Integer.valueOf(paramEntryValues[1]));
			setFlags(Flags.read(paramEntryValues[2]));
			setName(paramEntryValues[3]);
			setType(paramEntryValues[4]);
			setHelp(paramEntryValues[6]);
		} catch (IndexOutOfBoundsException e) {
			/** everything is ok, parameters have just finished*/
		} catch (NumberFormatException ex) {
			LOGGER.warn("wrong format of entry: " + line
					+ ", omitting");
			return null;
		}
		return build();
	}

	public void setField(String key, String value) {
		if (key.equals(ID_FIELD)) {
			setId(value);
		} else if (key.equals(NAME_FIELD)) {
			setName(value);
		} else if (key.equals(TYPE_FIELD)) {
			setType(value);
		} else if (key.equals(FLAGS_FIELD)) {
			setFlags(Flags.read(value));
		} else if (key.equals(HELP_FIELD)) {
			setHelp(value);
		} else if (key.equals(GROUP_FIELD)) {
			setGroup(Integer.valueOf(value));
		}
	}


	public static FramsClass getFramsClass() {
		return new FramsClass("prop", "prop", null)
				.append(new ParamBuilder().setId("name").setName("Name").setType(StringParam.class).build())
				.append(new ParamBuilder().setId("id").setName("Id").setType(StringParam.class).build())
				.append(new ParamBuilder().setId("type").setName("Type").setType(StringParam.class).build())
				.append(new ParamBuilder().setId("help").setName("Help").setType(StringParam.class).build())
				.append(new ParamBuilder().setId("flags").setName("Flags").setType(DecimalParam.class).build());
	}

}
